/*
    Agent - Event Scheduling  (AES) Library 
    Copyright (C) 2017  Marko Djogatovic

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

// Zaglavlje aes.h
#ifndef __AES_H__
#define __AES_H__

#include <iostream>
#include <string>
#include <cstdlib>
#include <cstring>

// Ukljucujemo prostor imena std;
using namespace std;

// Prostor imena esa
namespace aes {
  // Konstante
  const int max_broj_uredu  = 10;
  const int max_broj_ulisti = 50;
  
  // Predefinisacemo tip double u vreme
  using vreme = double; // alternativno, typedef double vreme;
  
  // Klasa agenta
  class agent {
  public:
    virtual ~agent() {}
  public:
    void posalji_poruku(const agent* ap,const string& poruka) {
      const_cast<agent*>(ap)->primi_poruku(this,poruka);
    }
    virtual void primi_poruku(const agent*,const string&) {};
  };
  // Apstraktna klasa dogadjaja
  class dogadjaj {
  protected:
    // Vreme nastupanja
    vreme   _vreme;
    // Agent
    agent** _agenti;
    // Broj agenata
    size_t  _broj;
  public:
    // Kontruktor
    dogadjaj(const size_t n):_vreme(0.0),_broj(n) {
      _agenti = new agent*[_broj];
    }
    // Destruktor
    virtual ~dogadjaj() {
      delete [] _agenti;
    }
    // Postavi agenta
    void postavi_agenta(const size_t n,const agent* a) {
      if(n<_broj)
        _agenti[n] = const_cast<agent*>(a);
      else {
        cerr << "dogadjaj::postavi_agenta - Pogresan indeks" << endl;
        exit(1);
      }
    }
    // Vrati agenta
    agent* vrati_agenta(const size_t n) const {
      if(n<_broj)
        return _agenti[n]; 
      else {
        cerr << "dogadjaj::vrati_agenta - Pogresan indeks" << endl;
        exit(1);
      }
      return nullptr;
    }
    // Postavi vreme nastupanja
    void postavi_vreme(const vreme v) {
      _vreme = v;
    }
    // Vrati vreme nastupanja
    vreme vrati_vreme() const { return _vreme; }
  public:
    // Cista virtuelna funkcija dogadjaja
    virtual void akcija() = 0;
  };
  // Klasa simulacija
  class simulacija {
  private:
    // Flag zaustavljanja
    bool _stop;
    // Broj entiteta u LBD listi
    size_t _broj;
    // Vreme simulacije
    vreme _vreme;
    // Polje entiteta u LBD listi
    dogadjaj* _lbd[max_broj_ulisti];
    // Sortiranje LBD - sortiranje spajanjem
    void sortiraj_lbd() {
      sortiranje_spajanjem(_lbd,0,_broj-1);
    }
    void sortiranje_spajanjem(dogadjaj* x[],const size_t l,const size_t r) {
      // Proverava se da li ima elemenata u listi koja se sortira
      if(l<r) {
        // Pivot elemet
        size_t m = (l+r)/2;
        // Sortira se deo polja do pivota
        sortiranje_spajanjem(x,l,m);
        // Sortira se deo polja od pivota
        sortiranje_spajanjem(x,m+1,r);
        // Spajaju se polja
        spajanje(x,l,m,r);
      }
    }
    void spajanje(dogadjaj* x[],const size_t l,const size_t m,const size_t r) {
      // Velicina pomocnih polja
      size_t nl = m-l+1, nd = r-m;
      // Pomocna polja
      dogadjaj **levo  = new dogadjaj*[nl];
      dogadjaj **desno = new dogadjaj*[nd];
      // Polja koja se spajaju se kopiraju u pomocna polja
      std::memcpy(levo ,x+l  ,nl*sizeof(dogadjaj*));
      std::memcpy(desno,x+m+1,nd*sizeof(dogadjaj*));
      // Smestamo elemente iz pomocnih polja u novo, sortirano polje
      size_t i,j,k;
      for(i=0,j=0,k=l; i<nl && j<nd; ++k) {
        // Poredjenje
        if(levo[i]->vrati_vreme()<=desno[j]->vrati_vreme())
          x[k] = levo[i++];
        else
          x[k] = desno[j++];
      }  
      // Preostali elementi se smestaju u novo polje
      if(i<nl) std::memcpy(x+k,levo+i ,(nl-i)*sizeof(dogadjaj*));
      else     std::memcpy(x+k,desno+j,(nd-j)*sizeof(dogadjaj*));
      // Brisemo pomocna polje
      delete [] levo;
      delete [] desno;
    }    
  public:
    // Konstruktor
    simulacija():_stop(false),_broj(0),_vreme(0.0) {}
    // Destruktor
    ~simulacija() {
      // Uklanjamo preostale entitete iz LBD na kraju simulacije.
      for(size_t i=0; i<_broj; ++i)
        delete _lbd[i];
    }
    // Rasporedjivanje dogadjaja
    void rasporedi(const dogadjaj* d) {
      if(_broj < max_broj_ulisti) {
        // Smestamo entitet u listu
        _lbd[_broj] = const_cast<dogadjaj*>(d);
        ++_broj;
        // Sortiramo LBD
        sortiraj_lbd();
      }
      else {
        cerr << "simulacija::rasporedi - Previse entiteta u listi dogadjaja" << endl;
        exit(1);
      }        
    }
    // Izvrsavanje simulacije
    void izvrsi() {
      // Flag zaustavljanja je netacan
      _stop = false;
      // Izvrsavamo simulaciju. Simulacija se zavrsava ukoliko
      // nema dogadjaja u LBD ili ukoliko je flag zaustavljanja tacan.
      do {
        // Vadimo prvi dogadjaj iz liste. Taj dogadjaj postaje tekuci dogadjaj.
        dogadjaj*_tekuci = _lbd[0];
        for(size_t i=1;i<_broj;++i)
          _lbd[i-1] = _lbd[i];
        --_broj;
        // Faza A: Azurirarmo vreme simulacije
        _vreme = _tekuci->vrati_vreme();
        // Faza B: Izvrsavamo dogadjaj
        _tekuci->akcija();
        // Brisemo tekuci dogadjaj iz memorije
        delete _tekuci;
      } while(_broj && !_stop);
    }
    // Zaustavljamo izvrsenje simulacije
    void zaustavi() { _stop = true; }
    // Vraca vreme simulacije
    vreme vrati_vreme() { return _vreme; }
  };
}

#endif // __AES_H__